# CHERRYTABLE CONNECTOR

The Cherrytable Connector allows the interaction with [Cherrytable](<http://www.cherry-data.com/>) through HTTP requests. Its functions permit to insert, read, update and delete rows through sequential, pipelined or queued requests.
The component is provided as a public python module which can be found at [PyPI](<https://pypi.org/project/cherrytable-connector/>) and can be installed in any machine with a version of **Python > 3**.

## Installation
The module can be installed with the following command
```
pip install cherrytable-connector
```

## Configuration
The connector object can be created as follows:

```python
from cherrytable_connector import cherrytable_connector as cc
connector = cc.CherrytableConnector(host, port, username, password, table_name, table_column_names, log_dir, debug, max_pipeline_requests)
```

where:
- _host_ and _port_ refer to the Cherrytable installation
- _username_ and _password_ are the credentials 
- _table\_name_ is the name of the table, e.g. “usertable”
- _table\_column\_names_  is the list of the column names, e.g. ["key\_column","field0","field1"] 
- _log\_dir_ is the path of the folder containing the logs, e.g. “../log”
- _debug_ is True to set the log in debug mode
- _max\_pipeline\_requests_ (optional parameter) is the maximum size of the pipeline

## Methods
### Sequential
Requests can be sequentially sent, i.e. waiting for the response of the first request before sending the second, using the `send` function: 
```python
send(req_type, key, data, options)
```

where:
- _req\_type_ is the type of the request
- _key_ is the value of the primary key of the row
- _data_ is the entire row, where values are divided by “;”
- _options_ is a list of column names, employed in the read function

Parameters vary according to the request type and examples of usage are the following:

#### INSERT
```python
res = connector.send(req_type=cc.RequestType.INSERT, data="1;33;172")
```

If the key (first value) already exists, the request won’t be successful. 

#### UPDATE
```python
res = connector.send(req_type=cc.RequestType.UPDATE, data="1;30;170")
```

To be successful the key should already exist.

#### READ
```python
res = connector.send(req_type=cc.RequestType.READ, key=”1”, options=[“key_column”, “field0”])
```

To be successful the key should already exist. With `options` it is possible to specify the columns to be retrieved, otherwise if `options` is None all the columns are retrieved. 

#### DELETE
```python
res = connector.send(req_type=cc.RequestType.DELETE, key=”1”)
```

To be successful the key should already exist.


The outcome `res` is a dictionary containing the fields `status` and `body`, corresponding to the status code of the response and its body.

### Pipelining

Pipelining allows to send multiple requests without waiting for individual responses. Requests are defined in a list and their responses are returned together. The maximum size of the pipeline can be specified in the constructor of CherrytableConnector. Requests are sent through the `sendPipeline` function:

```python
sendPipeline(requests)
```

which takes as input a list of requests, defined as tuples _(req\_type,key,data,options)_

```python
requests = [(cc.RequestType.INSERT, None, "1;33;172", None),\
            (cc.RequestType.INSERT, None, "2;33;172", None),\
            (cc.RequestType.INSERT, None, "3;33;172", None),\
            (cc.RequestType.DELETE, "1", None, None),\
            (cc.RequestType.DELETE, "2", None, None),\
            (cc.RequestType.UPDATE, None, "3;20;100", None),\
            (cc.RequestType.READ, "3", None, [“key_column”,”field0”]),\
            (cc.RequestType.DELETE, "3", None, None)]  
responses = connector.sendPipeline(requests)
```

`responses` is the list of all the response dictionaries, with `status` and `body` fields.

### Queued

With this feature it is possible to send requests without waiting for their responses. After having sent some requests it is possible to perform other operations before retrieving their responses.
Requests are automatically added to an internal queue and handled as pipelines. 
The main difference with respect to the prior method is that it is not necessary to prepare all the requests before sending them, instead requests are individually sent using:

```python
sendWithQueue(req_type, key, data, options)
```

For example:

```python    
requests = []
requests.append(connector.sendWithQueue(req_type=cc.RequestType.INSERT, key=None, data="1;33;172", options=None))
requests.append(connector.sendWithQueue(req_type=cc.RequestType.INSERT, key=None, data="2;33;172", options=None))
requests.append(connector.sendWithQueue(req_type=cc.RequestType.INSERT, key=None, data="3;33;172", options=None))
requests.append(connector.sendWithQueue(req_type=cc.RequestType.DELETE, key="1", data=None, options=None))

#Other operations…

responses=[]
last_completed = 0
while last_completed != (len(requests)-1):
if requests[last_completed+1].getResult()[0] == cc.RequestStatus.COMPLETED or requests[last_completed+1].getResult()[0] == cc.RequestStatus.ERROR:
               	last_completed +=1 
		responses.append(requests[last_completed+1].res) 
```

## Working with objects

Instead of using plain text to represent rows, it is possible to map rows to objects. 
For example, a Cherrytable row composed of the fields `key_column, field0, field1` can be mapped to a runtime object having the following fields:
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;c_id &rightarrow; key_column
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;age &rightarrow; field0
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;height &rightarrow; field1

In this way, object fields are updated according to the operations performed on the table.
```python
from cherrytable_connector import cherrytable_connector_obj as cc
connector = cc.CherrytableConnector(host, port, username, password, table_name, table_column_names, obj_mapping, log_dir, debug)
```

where `obj_mapping` is a list of fields of the object mapped positionally to `table_column_names`, in this example ["c_id","age","height"]

The object class must inherit from `cc.CherryObject` and can be defined for example as follows:

```python
class Customer(cc.CherryObject):
    def __init__(self, c_id, età, altezza):
        cc.CherryObject.__init__(self)
        self.c_id = c_id
        self.age = age
        self.height = height

customer = Customer(c_id="1", age="33", height="172")
```

### Sequential
Sequential requests are executed by sending the object as parameter:
```python
send(req_type, obj, options=None)
```
where 
- _req\_type_ is the type of the request
- _obj_ is the object corresponding to the table
- _options_ is a list of column names, employed in the read function

Parameters vary according to the request type and examples of usage are the following:

#### INSERT
```python
res = connector.send(req_type=cc.RequestType.INSERT, obj=customer)
```

If the key of the object (parameter corresponding to the primary key of the table) already exists, the request won’t be successful. 

#### UPDATE
```python
res = connector.send(req_type=cc.RequestType.UPDATE, obj=customer)
```

To be successful the key of the object should already exist.

#### READ
```python
res = connector.send(req_type=cc.RequestType.READ, obj=customer, options=[“key_column”, “field0”])
```

To be successful the key of the object should already exist. With `options` it is possible to specify the columns to be retrieved, otherwise if `options` is None all the columns are retrieved. 
When a read operation is performed, the object parameters are updated with the values contained in the corresponding table.

#### DELETE
```python
res = connector.send(req_type=cc.RequestType.DELETE, obj=customer)
```

To be successful the key of the object should already exist.


### Pipelining

The list of requests contained in a pipeline are defined as tuples _(req_type,object,options)_

```python
requests = [(cc.RequestType.INSERT,customer,None),\
            (cc.RequestType.INSERT,customer,None),\
            (cc.RequestType.INSERT,customer,None),\
            (cc.RequestType.DELETE,customer,None),\
            (cc.RequestType.DELETE,customer,None),\
            (cc.RequestType.UPDATE,customer_modified,None),\
            (cc.RequestType.READ,customer,None),\
            (cc.RequestType.DELETE,customer,None)]  
            
responses = connector.sendPipeline(requests)
```