'''
	Copyright (C) 2019  Cherrydata (http://www.cherry-data.com/)

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, see <https://www.gnu.org/licenses/>.
'''

from enum import Enum
from dugong import HTTPConnection
from queue import Queue

import urllib
import os
import logging
import logging.handlers
import threading

class RequestType(Enum):
	READ="read"
	INSERT="insert"
	UPDATE="update"
	DELETE="delete"

class RequestStatus(Enum):
	PENDING=0
	COMPLETED=1
	ERROR=2

class CherryRequest:
	def __init__(self,req_type,key=None,data=None,options=None):
		if not req_type:
			self.logger.error('Missing request type')
			raise ValueError('Missing request type')

		if not any(t for t in RequestType if t == req_type):
			raise TypeError('Invalid request type')

		self.req_type = req_type
		self.key = key
		self.data = data
		self.options = options

class CherryRequestWithQueue(CherryRequest):
	def __init__(self,req_type,key=None,data=None,options=None):
		self.lock = threading.RLock()
		self.status = RequestStatus.PENDING
		self.res = None
		CherryRequest.__init__(self,req_type,key,data,options)

	def __setattr__(self, attr, value):
		if attr in ('req_type','key','data','options') and hasattr(self, attr):
			raise Exception("Attempting to alter read-only value")
		
		if attr=='lock':
			self.__dict__[attr] = value
		else:
			self.lock.acquire()
			try:
				self.__dict__[attr] = value
			finally:
				self.lock.release()

	def getResult(self):
		return (self.status, self.res)

#One for each table
class CherrytableConnector:
	def __init__(self, host, port, username, password, table_name, table_column_names, log_dir, debug, log_id=0, max_pipeline_requests=10, max_queue_size=100):
		if not host or not port or not log_dir or not debug or not table_name or not table_column_names or not username or not password:
			raise ValueError('Missing one or more arguments')

		self.conn = HTTPConnection(hostname=host, port=port)
		self.username = username
		self.password = password

		self.table_name = table_name
		self.table_key_column_name = table_column_names[0]
		self.table_column_names = table_column_names

		self.max_pipeline_requests = max_pipeline_requests

		self.log_dir = log_dir
		self.log_id = log_id
		self.debug = debug
		self.setUpLogging(log_dir=log_dir, debug=debug)

		self.queue = Queue(maxsize=max_queue_size)
		thread = threading.Thread(target=self.__consumer)
		thread.start()

	def setUpLogging(self, log_dir, debug):
		log_file = os.path.join(log_dir, 'log_'+str(self.log_id)+'_cherrytable_connector_'+str(self.table_name)+'.txt')
		level = logging.INFO if debug else logging.WARNINGs
		handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=1000000, backupCount=5, mode='a')
		handler.setLevel(level)
		formatter = logging.Formatter('%(asctime)s - %(message)s')
		handler.setFormatter(formatter)
		self.logger = logging.getLogger(__name__)
		self.logger.setLevel(level)
		self.logger.addHandler(handler)

	def send(self,req_type,key=None,data=None,options=None):
		if not req_type:
			self.logger.error(str(self.log_id)+' - Missing request type')
			raise ValueError('Missing request type')
		if req_type == RequestType.READ:
			self.__read(key,options)
			return self.__processResponse(req_type)
		elif req_type == RequestType.INSERT:
			self.__insert(data)
			return self.__processResponse(req_type)
		elif req_type == RequestType.UPDATE:
			self.__update(data)
			return self.__processResponse(req_type)
		elif req_type == RequestType.DELETE:
			self.__delete(key)
			return self.__processResponse(req_type)
		else:
			self.logger.error(str(self.log_id)+' - Invalid request type')
			raise TypeError('Invalid request type')

	def sendWithQueue(self,req_type,key=None,data=None,options=None):
		req = CherryRequestWithQueue(req_type,key,data,options)
		self.queue.put(item=req,block=True)
		return req

	def sendPipeline(self,requests):
		if len(requests) > self.max_pipeline_requests:
			self.logger.error(str(self.log_id)+' - Exceeded pipeline size')
			raise ValueError('Exceeded pipeline size')
		
		requests_int = []
		for req in requests:
			cur_req = CherryRequest(req_type=req[0],key=req[1],data=req[2],options=req[3])
			requests_int.append(cur_req)

		thread = threading.Thread(target=self.__processRequests, args=(requests_int,))
		thread.start()
		thread.join()

		responses = []
		for req in requests_int:
			response = self.conn.read_response()
			body = self.conn.readall()
			body_text = body.decode('ascii')
			res = {}
			res['status'] = response.status
			res['body'] = body_text
			responses.append(res)
			if response.status == 200:
				self.logger.info(str(self.log_id)+' - ['+str(req.req_type.value).upper()+'] SUCCESS')
				if req.req_type == RequestType.READ:
					self.logger.info(str(self.log_id)+' - RESPONSE BODY: ' + str(body_text))
			else:
				self.logger.error(str(self.log_id)+' - ['+str(req.req_type.value).upper()+'] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' + str(body_text))		
		return responses
			
	#options must be consecutive and containing the keyvalue, e.g. ['key_column', 'field0']
	def __read(self,key,options=None):
		if not key:
			self.logger.error(str(self.log_id)+' - [READ] Missing parameter: key')
			raise ValueError('[READ] Missing parameter: key')

		url_dict = { 'action':'table_read_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'key_value': key }
		if options:
			fields = str(options[0])
			for i in range(1, len(options)):
				fields += ',' + str(options[i])
			url_dict.update({'fields': fields})
			
		headers = {'Connection' : 'keep-alive','Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)

		self.logger.info(str(self.log_id)+' - [READ] QUERY: ' + url)
		self.conn.send_request(method="GET", path=url, headers=headers)
	
	def __delete(self,key):
		if not key:
			self.logger.error(str(self.log_id)+' - [DELETE] Missing parameter: key')
			raise ValueError('[DELETE] Missing parameter: key')

		url_dict = { 'action':'table_delete_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'key_value': key }
		headers = {'Connection' : 'keep-alive', 'Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)

		self.logger.info(str(self.log_id)+' - [DELETE] QUERY: ' + url)
		self.conn.send_request(method="DELETE", path=url, headers=headers)		#TODO:delete

	#data="k;f0;f1;f2;f3;f4;f5"
	def __insert(self,data):
		if not data:
			self.logger.error(str(self.log_id)+' - [INSERT] Missing parameter: data')
			raise ValueError('[INSERT] Missing parameter: data')

		url_dict = { 'action':'table_insert_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'imode': 'reject_multiple_records_per_key' }
		headers = { 'Content-Type': 'text/plain', 'Connection' : 'keep-alive', 'Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)
		
		self.logger.info(str(self.log_id)+' - [INSERT] QUERY: ' + url)
		self.conn.send_request(method="PUT", path=url, body=data.encode('ascii'), headers=headers)
	
	#data = complete row
	def __update(self,data):
		if not data:
			self.logger.error(str(self.log_id)+' - [UPDATE] Missing parameter: data')
			raise ValueError('[UPDATE] Missing parameter: data')

		url_dict = { 'action':'table_insert_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'imode': 'drop_older_record_with_same_key' }
		headers = { 'Content-Type': 'text/plain', 'Connection' : 'keep-alive', 'Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)

		self.logger.info(str(self.log_id)+' - [UPDATE] QUERY: ' + url)
		self.conn.send_request(method="PUT", path=url, body=data.encode('ascii'), headers=headers)

	def __processRequests(self,requests):
		self.logger.info(str(self.log_id)+' - STARTED PIPELINE with '+str(len(requests))+' requests')

		for req in requests:
			self.logger.info(str(self.log_id)+' - ->' + str((req.req_type.value).upper()))
			if req.req_type == RequestType.READ:
				self.__read(req.key, req.options)
			elif req.req_type == RequestType.INSERT:	
				self.__insert(req.data)
			elif req.req_type == RequestType.UPDATE:
				self.__update(req.data)
			elif req.req_type == RequestType.DELETE:
				self.__delete(req.key)

	def __processResponse(self,req_type):
		response = self.conn.read_response()
		body = self.conn.readall()
		body_text = body.decode('ascii')
		res = {}
		res['status'] = response.status
		res['body'] = body_text
		if req_type == RequestType.READ:
			if response.status == 200:
				self.logger.info(str(self.log_id)+' - [READ] SUCCESS - RESPONSE BODY: ' + str(body_text))
			else:
				self.logger.error(str(self.log_id)+' - [READ] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' +str(body_text))
			return res
		elif req_type == RequestType.INSERT or req_type == RequestType.UPDATE or req_type == RequestType.DELETE:
			if response.status == 200:
				self.logger.info(str(self.log_id)+' - ['+str(req_type.value).upper()+'] SUCCESS')
			else:
				self.logger.error(str(self.log_id)+' - ['+str(req_type.value).upper()+'] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' + str(body_text))		
			return res

	def __consumer(self):
		while True:
			reqs = []
			counter = self.max_pipeline_requests
			while not self.queue.empty() and counter>0:
				reqs.append(self.queue.get())
				counter-=1

			if reqs:
				self.logger.info(str(self.log_id)+' - CONSUMER: PROCESSING ' + str(len(reqs)) + " REQUESTS")
				thread = threading.Thread(target=self.__processRequests, args=(reqs,))
				thread.start()
				thread.join()

				for req in reqs:
					response = self.conn.read_response()
					body = self.conn.readall()
					body_text = body.decode('ascii')
					result = {}
					result['status'] = response.status
					result['body'] = body_text
					if response.status == 200:
						self.logger.info(str(self.log_id)+' - ['+str(req.req_type.value).upper()+'] SUCCESS')
						if req.req_type == RequestType.READ:
							self.logger.info(str(self.log_id)+' - RESPONSE BODY: ' + str(body_text))
						req.status = RequestStatus.COMPLETED
						req.res = result
					else:
						self.logger.error(str(self.log_id)+' - ['+str(req.req_type.value).upper()+'] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' + str(body_text))		
						req.status = RequestStatus.ERROR