'''
    Copyright (C) 2019  Cherrydata (http://www.cherry-data.com/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <https://www.gnu.org/licenses/>.
'''

from enum import Enum
from dugong import HTTPConnection

import urllib
import os
import logging
import logging.handlers
import threading

class RequestType(Enum):
	READ="read"
	INSERT="insert"
	UPDATE="update"
	DELETE="delete"

class CherryObject:
	def __init__(self):
		self.lock = threading.RLock()

	def __setattr__(self, attr, value):
		if attr=='lock':
			self.__dict__[attr] = value
		else:
			self.lock.acquire()
			try:
				self.__dict__[attr] = value
			finally:
				self.lock.release()

class CherryRequest:
	def __init__(self,req_type,obj,options=None):
		if not req_type or not obj:
			self.logger.error('Missing request parameters')
			raise ValueError('Missing request parameters')

		if not any(t for t in RequestType if t == req_type):
			raise TypeError('Invalid request type')

		self.req_type = req_type
		self.obj = obj
		self.options = options

#One for each table
class CherrytableConnector:
	def __init__(self, host, port, table_name, table_column_names, obj_mapping, log_dir, debug, username, password, max_pipeline_requests=10, max_queue_size=100):
		if not host or not port or not log_dir or not debug or not table_name or not table_column_names or not obj_mapping or not username or not password:
			raise ValueError('Missing one or more arguments')
		if len(table_column_names) != len(obj_mapping):
			raise ValueError('Invalid object mapping argument')

		self.conn = HTTPConnection(hostname=host, port=port)
		self.username = username
		self.password = password

		self.table_name = table_name
		self.table_key_column_name = table_column_names[0]
		self.table_column_names = table_column_names
		self.obj_key_column_name = obj_mapping[0]
		self.obj_mapping = obj_mapping

		self.max_pipeline_requests = max_pipeline_requests

		self.log_dir = log_dir
		self.debug = debug
		self.setUpLogging(log_dir=log_dir, debug=debug)

	def setUpLogging(self, log_dir, debug):
		log_file = os.path.join(log_dir, 'log_cherrytable_connector_obj_'+str(self.table_name)+'.txt')
		level = logging.INFO if debug else logging.WARNING
		handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=1000000, backupCount=5, mode='a')
		handler.setLevel(level)
		formatter = logging.Formatter('%(asctime)s - %(message)s')
		handler.setFormatter(formatter)
		self.logger = logging.getLogger(__name__)
		self.logger.setLevel(level)
		self.logger.addHandler(handler)

	def send(self,req_type,obj,options=None):
		if not req_type or not obj:
			self.logger.error('Missing parameters')
			raise ValueError('Missing parameters')
		if req_type == RequestType.READ:
			self.__read(obj)
			return self.__processResponse(req_type, obj)
		elif req_type == RequestType.INSERT:
			self.__insert(obj)
			return self.__processResponse(req_type, obj)
		elif req_type == RequestType.UPDATE:
			self.__update(obj)
			return self.__processResponse(req_type, obj)
		elif req_type == RequestType.DELETE:
			self.__delete(obj)
			return self.__processResponse(req_type, obj)
		else:
			self.logger.error('Invalid request type')
			raise TypeError('Invalid request type')

	def sendPipeline(self,requests):
		if len(requests) > self.max_pipeline_requests:
			self.logger.error('Exceeded pipeline size')
			raise ValueError('Exceeded pipeline size')
		
		requests_int = []
		for req in requests:
			cur_req = CherryRequest(req_type=req[0],obj=req[1],options=req[2])
			requests_int.append(cur_req)

		thread = threading.Thread(target=self.__processRequests, args=(requests_int,))
		thread.start()
		thread.join()

		responses = []
		for req in requests_int:
			response = self.conn.read_response()
			body = self.conn.readall()
			body_text = body.decode('ascii')
			res = {}
			res['status'] = response.status
			res['body'] = body_text
			responses.append(res)
			if response.status == 200:
				self.logger.info('['+str(req.req_type.value).upper()+'] SUCCESS')
				if req.req_type == RequestType.READ:
					values = body_text.split(';')	
					for i in range(1, len(values)):			#skip keyColumn value
						setattr(req.obj, self.obj_mapping[i], values[i])
					self.logger.info("RESPONSE BODY: " + str(body_text))
			else:
				self.logger.error('['+str(req.req_type.value).upper()+'] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' + str(body_text))		
		return responses

	#options must be consecutive and containing the keyvalue, e.g. ['key_column', 'field0']
	def __read(self,obj,options=None):
		if not obj:
			self.logger.error('[READ] Missing object')
			raise ValueError('[READ] Missing object')

		key = getattr(obj, self.obj_key_column_name)
		
		url_dict = { 'action':'table_read_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'key_value': key }
		if options:
			fields =  str(options[0])
			for i in range(1, len(options)):
				fields += ',' + str(options[i])
			url_dict.update({'fields': fields})
		headers = {'Connection' : 'keep-alive','Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)

		self.logger.info('[READ] QUERY: ' + url)
		self.conn.send_request(method="GET", path=url, headers=headers)		
	
	def __delete(self, obj):
		if not obj:
			self.logger.error('[DELETE] Missing object')
			raise ValueError('[DELETE] Missing object')
		
		key = getattr(obj, self.obj_key_column_name)
		
		url_dict = { 'action':'table_delete_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'key_value': key }
		headers = {'Connection' : 'keep-alive', 'Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)

		self.logger.info('[DELETE] QUERY: ' + url)
		self.conn.send_request(method="GET", path=url, headers=headers)

	def __insert(self, obj):
		if not obj:
			self.logger.error('[INSERT] Missing object')
			raise ValueError('[INSERT] Missing object')
		
		url_dict = { 'action':'table_insert_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'imode': 'reject_multiple_records_per_key' }
		headers = { 'Content-Type': 'text/plain', 'Connection' : 'keep-alive', 'Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)
		
		values = []
		for field_name in self.obj_mapping:
			values.append(getattr(obj, field_name))
		data = ';'.join(values)

		self.logger.info('[INSERT] QUERY: ' + url)
		self.conn.send_request(method="PUT", path=url, body=data.encode('ascii'), headers=headers)

	#data = complete row
	def __update(self, obj):
		if not obj:
			self.logger.error('[UPDATE] Missing object')
			raise ValueError('[UPDATE] Missing object')
		
		url_dict = { 'action':'table_insert_row',
					 'table_name': self.table_name, 
					 'key_column_name': self.table_key_column_name,
					 'imode': 'drop_older_record_with_same_key' }
		headers = { 'Content-Type': 'text/plain', 'Connection' : 'keep-alive', 'Authorization' : 'Basic %s' % self.username+":"+self.password}
		url = '/ycsb?' + urllib.parse.urlencode(url_dict)

		values = []
		for field_name in self.obj_mapping:
			values.append(getattr(obj, field_name))
		data = ';'.join(values)
		
		self.logger.info('[UPDATE] QUERY: ' + url)
		self.conn.send_request(method="PUT", path=url, body=data.encode('ascii'), headers=headers)

	def __processRequests(self,requests):
		self.logger.info('STARTED PIPELINE with '+str(len(requests))+' requests')

		for req in requests:
			self.logger.info("->" + str((req.req_type.value).upper()))
			if req.req_type == RequestType.READ:
				self.__read(req.obj,req.options)
			elif req.req_type == RequestType.INSERT:	
				self.__insert(req.obj)
			elif req.req_type == RequestType.UPDATE:
				self.__update(req.obj)
			elif req.req_type == RequestType.DELETE:
				self.__delete(req.obj)

	def __processResponse(self,req_type,obj):
		response = self.conn.read_response()
		body = self.conn.readall()
		body_text = body.decode('ascii')
		res = {}
		res['status'] = response.status
		res['body'] = body_text
		if req_type == RequestType.READ:	
			if response.status == 200:
				values = body_text.split(';')	
				for i in range(1, len(values)):			#skip keyColumn value
					setattr(obj, self.obj_mapping[i], values[i])
				self.logger.info('[READ] SUCCESS - RESPONSE BODY: ' + str(body_text))
			else:
				self.logger.error('[READ] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' +str(body_text))
			return res
		elif req_type == RequestType.INSERT or req_type == RequestType.UPDATE or req_type == RequestType.DELETE:
			if response.status == 200:
				self.logger.info('['+str(req_type.value).upper()+'] SUCCESS')
			else:
				self.logger.error('['+str(req_type.value).upper()+'] FAILED WITH STATUS CODE ' + str(response.status) + ' - ' + str(response.reason) + ' - ' + str(body_text))		
			return res